package simple.dcsp.abtfamily.abtdac;

import java.util.Iterator;

import kernel.Communicator;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.events.AgentEventListener;
import kernel.events.VariableEvent;
import kernel.order.AgentOrdering;
import kernel.vars.integer.IntExternalVar;
import kernel.vars.integer.explain.ExplainedIntVar;
import simple.Nogood;
import simple.SingleExplainedVarAgent;
import simple.dcsp.abtfamily.ABT;
import simple.dcsp.abtfamily.IdleEvent;
import simple.dcsp.abtfamily.abt.ABTTermDetection;
import simple.dcsp.abtfamily.abt.AddLinkMsgHandler;
import simulation.metric.Metric;
import util.Commands;
import util.IntArray;
import util.VarAssignment;
import util.VarEventMask;

public class ABTdacAgent extends SingleExplainedVarAgent implements ABT {

	protected OkMessageHandler okMsgH;
	protected NogoodsHandler ngdMsgH;
	protected DelMsgHandler delMsgH;
	protected AddLinkMsgHandler adlMsgH;

	// =================================================================================

	/**
	 * @param agentId
	 * @param agName
	 * @param description
	 */
	public ABTdacAgent(int agentId, String agName, String description) {
		super(agentId, agName, description);
		this.agentOrder = new AgentOrdering(this);
		this.terminateMsgHandler = new ABTTermDetection(this);
		this.constraintHandler = new ABTdacConstraintHandler(this);
		this.okMsgH = new OkMessageHandler(this);
		this.ngdMsgH = new NogoodsHandler(this);
		this.adlMsgH = new AddLinkMsgHandler(this);
		this.delMsgH = new DelMsgHandler(this);
		this.constructCommunicator();
		this.initAuxStatistics();
	}

	private void initAuxStatistics() {
		this.auxiliaryStatistics.put(Commands.nbObsoleteNogood, new Metric(Commands.nbObsoleteNogood));
		this.auxiliaryStatistics.put(Commands.nbAssignments, new Metric(Commands.nbAssignments));
		this.auxiliaryStatistics.put(Commands.ObsoleteOk, new Metric(Commands.ObsoleteOk));
	}

	private void constructCommunicator() {
		this.communicator = new Communicator(this) {
			protected boolean waitForArrivalMessages(long msTimeout) throws InterruptedException {
				logDomains();
				this.waitingForMessage = true;
				boolean arrived = false;
				if (((ABT) this.myAgent).isIdle()) {
					arrived = mailBoxLock.waitUntilTrue(10);
					if (!arrived)
						this.myAgent.fireAgentEvent(new IdleEvent(this.myAgent));
				}
				if (!arrived) {
					arrived = mailBoxLock.waitUntilTrue(msTimeout);
				}
				this.waitingForMessage = false;
				logDomains();
				return arrived;
			}
		};
	}

	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.DESINSTANCIATE);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_THE_LAST).add(VariableEvent.RESTORE_VAL);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_BETWEEN_L_H).add(VariableEvent.RESTORE_VAL);

		// cette ligne est pour DAC
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_THE_LAST).add(VariableEvent.REMOVE_VAL);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_LOWER_THAN_LOCAL_VAR).add(VariableEvent.REMOVE_VAL);

		this.constraintHandler.update();
	}

	// =================================================================================

	@Override
	public void addMsgHandlers() {
		this.communicator.addMsgHandler(this.okMsgH);
		this.communicator.addMsgHandler(this.ngdMsgH);
		this.communicator.addMsgHandler(this.adlMsgH);
		this.communicator.addMsgHandler(this.delMsgH);
	}

	// ==========================================================================

	public void initAgentView() {
		IntArray agents = this.agentOrder.getHigherAgents();
		for (int agId : agents) {
			if (this.communicator.isNeighbor(agId)) {
				this.agentView.add((IntExternalVar) this.getExternalVar(agId));
			}
		}
	}

	// =================================================================================
	// ABT Procedures
	// =================================================================================
	@Override
	public void startDistributedSearch() throws StopException {
		this.addAgentEventListener((AgentEventListener) this.terminateMsgHandler);
		((ABTTermDetection) this.terminateMsgHandler).sendInitialEndAgentMessage();

		if (this.getMailBoxLock().isTrue())
			this.processSomeMessages();

		this.checkAgentView();
	}

	// =================================================================================
	/**
	 * choose an instantiation (i.e a value) compatible with the agentView
	 * 
	 * @throws StopException
	 */
	public void checkAgentView() throws StopException {
		int myVal = this.var.getValue();
		if (!this.var.getDomain().contains(myVal)) {
			if (this.valSelector.hasNextValue()) {
				int val = this.valSelector.getNextValue();
				this.var.instantiate(val);
				if (var.getPreviousValue() != var.getValue())
					this.okMsgH.sendOkMessages();
			} else {
				this.backtrack();
			}
		}
	}

	// =================================================================================
	public boolean isIdle() {
		if (this.disSearchStarted && this.var.isInstantiated()) {
			IntArray hn = communicator.getHigherNeighbors();
			for (int id : hn) {
				IntExternalVar v = this.agentView.getVar(new VarIdentifier(id));
				if (!v.isInstantiated())
					return false;
			}
			return true;
		}

		return false;
	}

	// =================================================================================
	/**
	 * @throws StopException
	 */
	public void backtrack() throws StopException {
		this.resolveContradiction();

		this.checkAgentView();
	}

	// =================================================================================

	public void resolveContradiction() throws StopException {
		Nogood nogood = this.solveMyNogoods();
		if (nogood.isEmpty()) { throw new StopException(Commands.FAILURE); }
		this.agentView.clear(nogood.getRhs());
		this.ngdMsgH.sendNogood(nogood);
	}

	// =================================================================================
	/**
	 * Procedure to check if a new link should be added.
	 * 
	 * @param nogood
	 */
	public void checkAddLink(Nogood nogood) {
		Iterator<VarAssignment> it = nogood.getLhs().iterator();
		while (it.hasNext()) {
			VarAssignment ass = it.next();
			if (!this.communicator.isHigherNeighbor(ass.getAgentId())) {
				this.communicator.addNeighbor(ass.getAgentId());
				this.agentView.add(this.makeExternalVar(ass));
				this.adlMsgH.sendAddLink(ass);
			}
		}
	}

	// =================================================================================

	public void valuesConditionallyDeleted(Nogood nogood) throws StopException {
		if (this.agentView.compatibleWith(nogood.getLhs())) {
			// IntExternalVar externalVar = (IntExternalVar) this.getExternalVar(nogood.getRhsAgentId());
			this.makeNoGoodConstraint(nogood);

			this.checkAddLink(nogood);
			this.checkAgentView();
		}

	}

	// =================================================================================

	public void sendConditionalDeldMsg(int val) {
		Nogood ng = this.makeBetterNogood((ExplainedIntVar) var, val);
		int last = ng.getLhs().last().getAgentId();
		IntArray neighbors0 = this.communicator.getHigherNeighbors0();
		for (int agId : neighbors0) {
			if (this.agentOrder.chooseHigher(agId, last) == last && last != agId && ng.getLhs().size() > 1) {
				this.delMsgH.sendConditionalDeldMsg(ng, agId);
			}
		}
	}

	// =================================================================================

	@Override
	public void sendOkMessage(int agId) {
		this.okMsgH.sendOkMessage(agId);
	}

}
