package simulation;

import java.sql.Date;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.jfree.data.category.DefaultCategoryDataset;

import kernel.AbstractAgent;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.order.CentralizedAO;
import simple.SimpleAgent;
import simulation.metric.CommunicationObserver;
import simulation.metric.MetricListener;
import simulation.metric.NCCCObserver;
import util.Commands;
import util.Instantiation;
import util.IntArray;
import util.Logger;
import util.VarAssignment;
import util.parser.PConstraint;
import util.parser.PExtensionConstraint;
import util.parser.PIntensionConstraint;
import util.parser.PVariable;

/**
 * A class that allows you to run and control the distributed search. It is used to solve instances of distributed problems when simulation.
 * 
 * @authors 
 * <a href="mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@lirmm.fr &cc=bessiere@lirmm.fr"> Redoaune Ezzahir</a> and 
 * <a href="mailto:wahbi@lirmm.fr?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr"> Mohamed Wahbi</a>
 * 
 */
public class DisSolverSimulator extends MailerAMDS {

	/** A field indicates to the solver if it keep the agents running or not. */
	private volatile boolean keepRunning = true;

	/** A field storing the test period of the termination */
	protected long checkingtime = 10L; // mSecond

	protected long maxLoopPrint = 1000;

	int numLoopRun = 0;
	/**
	 * All thread agents controlled by the solver
	 */
//	final Thread[] threads;
	/**
	 * A list of IDs of registered agents (the agents of problem)
	 */
	private IntArray registredAgentIds;

	private boolean checkTheReturnedSolution;

	private DefaultCategoryDataset dataSetCCs;

	private double cpuTime = Long.MIN_VALUE; 

	private double comulatedCpuTime = 0;

	private Thread masterThread;

	/**
	 * Constructs a new distributed solver simulator
	 * 
	 * @param disProb
	 *            the problem to solve
	 */
	public DisSolverSimulator(DisProblem disProb) {
		super(disProb);
		this.master = disProb.master;
		this.master.log("solving problem: "+disProb.getProblemName()+" by :"+this.master.getClass(), Logger.INFO_LEVEL);
		this.registredAgentIds = new IntArray();
//		this.threads = new Thread[(disProb.getNumAgents() + 1)];// master and
		this.dataSetCCs = new DefaultCategoryDataset(); // all agents
	}

	/**
	 * Called to stop running of the solver
	 */
	public void stoptRunning() {
		this.keepRunning = false;
	}

	/**
	 * to solve the problem
	 */
	public void solve() {
		// this.addNCCCsObservers();
		this.setMailerAsCommunicationInterface();
		this.run();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.CommunicationInterface#register(int)
	 */
	public void register(int agentId) {
		this.registredAgentIds.add(agentId);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.CommunicationInterface#isRegistered(int)
	 */
	public boolean isRegistered(int agentId) {
		return this.registredAgentIds.contains(agentId);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.CommunicationInterface#getRegisteredAgentIds()
	 */
	@Override
	public IntArray getRegisteredAgentIds() {
		return this.registredAgentIds;
	}
	
	/**
	 * Called by {@link #solve()} method
	 */
	public void run() {
		this.startThreads();
		Thread.yield();
		int loop = 0;
		while (this.keepRunning) {
			try {
				Thread.sleep(this.checkingtime);
			} catch (InterruptedException e) {
				if (Logger.LOG_DEBUG) {
					this.master.log(e.toString(), Logger.ERROR_LEVEL);
				}
				e.printStackTrace();
				System.exit(-1);
			}
			if (this.evaluateMessageDelay && this.checkStillness()) {
				if (this.whenStillness() == 0) {
					System.err.println(
							"Warnings : All agents are in waiting state! (" + numLoopRun + ") "
							+ "\n I will terminate this instance at (20) NCCT: "+this.NCCT);
					numLoopRun++;
				} else {
					numLoopRun = 0;
				}
			}
			this.keepRunning = !this.checkEnd();
			loop++;
			if (loop > this.maxLoopPrint) {
				if (!this.evaluateMessageDelay && this.checkWait()) {
					System.err.println(
							"Warnings : All agents are in waiting state! (" + numLoopRun + ") \n" +
							"I will terminate this instance at (20) NCCT: "+this.NCCT);
					numLoopRun++;
				} else {
					numLoopRun = 0;
				}
				loop = 0;
				System.gc();
			}
			if (numLoopRun == 20) {
				this.Interrupt();
			}
		}
	}
	
	/**
	 * 
	 */
	private void Interrupt() {
		HashMap<Integer, AbstractAgent> agents = this.problem.getAgents();
		Iterator<Integer> it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = agents.get(it.next());
			ag.interrupt();
		}
		this.master.interrupt();
		try {
			Thread.sleep(20 * agents.size());
		} catch (InterruptedException e) {
			if (Logger.LOG_DEBUG) {
				this.master.log(e.toString(), Logger.ERROR_LEVEL);
			}
			e.printStackTrace();
			System.exit(-1);
		}
		this.keepRunning = false;
	}
	
	/**
	 * 
	 * @return
	 */
	private boolean checkWait() {
		if (!this.master.isWaitingForMessage())
			return false;
		HashMap<Integer, AbstractAgent> agents = this.problem.getAgents();
		Iterator<Integer> it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = agents.get(it.next());
			if (!ag.isWaitingForMessage())
				return false;
		}
		return true;
	}

	/**
	 * Start all agents threads
	 */

	private void startThreads() {
//		this.bean = ManagementFactory.getThreadMXBean();
//		if (!bean.isThreadCpuTimeSupported()) {
//			System.out.println("Sorrt, the Java virtual machine implementation " +
//					"does not support CPU time measurement for any thread.");
//			System.exit(-1);
//		}
//		if (!bean.isCurrentThreadCpuTimeSupported()) {
//			System.out.println("Sorry, JVM does not supports CPU time measurement " +
//					"for the current thread");
//			System.exit(-1);
//		}
//		if (!bean.isThreadCpuTimeEnabled()) {
//			System.out.println("Sorry, Thread CPU measurement is not enabled");
//			System.exit(-1);
//		}
//		int i = 0;
//		this.threads[i] = new Thread(this.master);
//		this.threads[i++].start();
		this.masterThread = new Thread(this.master);
		this.masterThread.start();
		HashMap<Integer, AbstractAgent> agents = this.problem.getAgents();
		Iterator<Integer> it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = agents.get(it.next());
			ag.start();
//			this.threads[i] = new Thread(ag);
//			this.threads[i++].start();
		}
	}

	@SuppressWarnings("deprecation")
	public void setThreadsToNull() {
//		for (int i = 0 ; i < this.threads.length ; i++) {
//			this.threads[i] = null;
//		}
		this.masterThread = null;
		HashMap<Integer, AbstractAgent> agents = this.problem.getAgents();
		Iterator<Integer> it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = agents.get(it.next());
			ag.stop();
			ag = null;
		}
	}

	/**
	 * Stop all agents threads
	 */
	public void stopThreads() {
		HashMap<Integer, AbstractAgent> agents = this.problem.getAgents();
		Iterator<Integer> it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = agents.get(it.next());
			try {
				ag.interrupt();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		try {
			this.masterThread.interrupt();
		} catch (Exception e) {
			e.printStackTrace();
		}
//		for (int i = 0 ; i < this.threads.length ; i++) {
//			try {
//				this.threads[i].interrupt();
//			} catch (Exception e) {
//				e.printStackTrace();
//			}
//		}
	}

	/**
	 * Suspend all agents threads
	 */
	@SuppressWarnings("deprecation")
	public void suspendThreads() {
//		for (int i = 0 ; i < this.threads.length ; i++) {
//			this.threads[i].suspend();
//		}
		HashMap<Integer, AbstractAgent> agents = this.problem.getAgents();
		Iterator<Integer> it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = agents.get(it.next());
			ag.suspend();
		}
		this.masterThread.suspend();
	}

	/**
	 * Suspend all agents threads
	 */
	@SuppressWarnings("deprecation")
	public void resumeThreads() {
//		for (int i = 0 ; i < this.threads.length ; i++) {
//			this.threads[i].resume();
//		}
		HashMap<Integer, AbstractAgent> agents = this.problem.getAgents();
		Iterator<Integer> it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = agents.get(it.next());
			ag.resume();
		}
		this.masterThread.resume();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.CommunicationInterface#checkEnd()
	 */
	public boolean checkEnd() {
		if (!this.master.isFinished()) return false;
		HashMap<Integer, AbstractAgent> agents = this.problem.getAgents();
		Iterator<Integer> it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = agents.get(it.next());
			if (!ag.isFinished()) return false;
		}
		it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = agents.get(it.next());
			double agCPUTime = ag.getCPUTime();
			this.comulatedCpuTime += agCPUTime;
			this.cpuTime = Math.max(this.cpuTime, agCPUTime);//TODO
//			System.out.println(ag.getAgentId()+" : "+(ag.getCPUTime() / 1.0E9));
		}
		return true;
	}


	/**
	 * Returns global solution in the form of string
	 * 
	 */
	@Override
	public String getGlobalSolution() {
		String solution = "";
		Iterator<AbstractAgent> it = this.getDisProblem().getAgents().values().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = it.next();
			solution += ag.solutionToString() + " ";
		}
		return solution;
	}

	public Comparator<? super VarAssignment> getComparator(final IntArray order) {
		return new Comparator<VarAssignment>() {
//			public int compare(Object paramT1, Object paramT2) {
//				VarAssignment v1 = ((VarAssignment) paramT1);
//				VarAssignment v2 = ((VarAssignment) paramT2);
//				return order.indexOf(v1.getAgentId()) - order.indexOf(v2.getAgentId());
//			}
			@Override
			public int compare(VarAssignment v1, VarAssignment v2) {
				return order.indexOf(v1.getAgentId()) - order.indexOf(v2.getAgentId());
			}
		};
	}

	public Instantiation getSolution() {
		IntArray order = this.master.getOrder();
		Instantiation solution = new Instantiation(this.getComparator(order));
		for (int agId : order) {
			AbstractAgent ag = this.problem.getAgent(agId);
			if (ag instanceof SimpleAgent) {
				solution.add(((SimpleAgent) ag).var.getVarAssignment());
			}
		}
		return solution;
	}

	public void addMetric(MetricListener metric) {
		this.master.addMetric(metric);
	}

	@Override
	public void addNCCCMetric() {
		this.master.addMetric(new NCCCObserver());
	}

	@Override
	public void setCentralizedAO(CentralizedAO centralizedAO) {
		this.master.setCentralizedAO(centralizedAO);
	}

	@Override
	public void addCommunicationMetric() {
		this.master.addMetric(new CommunicationObserver());
	}

	@Override
	public HashMap<String, Double> getStatistics() {
		if (this.checkTheReturnedSolution) {
			if (this.getEndState().equals(Commands.SUCCESS)) {
				String pb = this.isSolution();
				if (pb != null) {
					Logger log = new Logger("Error_"+new Date(0)+".txt");
					log.printToLogln("trying to solve the problem :\n"+this.problem.getProblemFile());
					log.printToLogln("using the protocol : "+this.problem.getProtocol());
					log.printToLogln("master : "+this.problem.master.getClass());
					log.printToLogln(pb);
					log.close();
					throw new Error(pb);
				}
			} else { }
		}
		HashMap<String, Double> statistics = this.master.getStatistics();
		statistics.put(Commands.CPU, (this.cpuTime)/1.0E9);//TODO
		statistics.put(Commands.comulatedCpuTime, this.comulatedCpuTime/1.0E9);
		statistics.put(Commands.Time, this.master.getTotalCPUTime()/((long)1e6));
		return statistics; 
	}
	
//	/**
//	 * @return
//	 */
//	private long getCPUTime() { //TODO
//		for(AbstractAgent ag : this.problem.getAgents().values()){
//			long cpu = ag.getCpuTime();
//			this.comulatedCpuTime  += cpu;
////			System.out.println("agId : "+ag.getAgentId()+" cpu : "+cpu/((long)1e6));
//			if(this.cpuTime < cpu){
//				this.cpuTime = cpu;
//			}
//		}
////		System.out.println("cpuTime : "+cpuTime);
//		return cpuTime;
//	}

	/**
	 * 
	 * @param check
	 */
	public void CheckTheReturnedSolution(boolean check) {
		this.checkTheReturnedSolution = check;
	}
	
	/**
	 * 
	 * @return
	 */
	public String isSolution() {
		Instantiation solution = this.getSolution();
		Map<String, PConstraint> constraints = this.problem.parser.getMapOfConstraints();
		Iterator<String> it = constraints.keySet().iterator();
		while (it.hasNext()) {
			PConstraint constraint = constraints.get(it.next());
			PVariable[] scope = constraint.getScope();
			int agId1 = scope[0].getAgent().getId();
			int agId2 = scope[1].getAgent().getId();
			int val1 = solution.getVarAssignment(new VarIdentifier(agId1)).getValue();
			int val2 = solution.getVarAssignment(new VarIdentifier(agId2)).getValue();
			if(constraint instanceof PIntensionConstraint) return null;
			int[][] tuples = ((PExtensionConstraint) constraint).getRelation().getTuples();
			for (int i = 0 ; i < tuples.length ; i++) {
				int v1 = tuples[i][0];
				int v2 = tuples[i][1];
				if (val1 == v1 && val2 == v2) {
					String s = "================================================";
					String problem = "\n" + s + s + s + s + "\n";
					problem += "The constraint " + constraint.getName() + " is not satisfied " + 
					" \t==>  " + agId1 + "=" + val1 + " <> " + agId2 + "="+ val2 + "\n";
					problem += solution + "\n" + s + s + s + s + "\n";
					AbstractAgent ag1 = this.getDisProblem().getAgent(agId1);
					AbstractAgent ag2 = this.getDisProblem().getAgent(agId2);
					problem += ag1.constraintHandler + "\n" ;
					problem += "AgentView: "+ag1.agentView+"\n"+ s + s + s + s + "\n";
					problem += ag2.constraintHandler + "\n" ;
					problem += "AgentView: "+ag2.agentView+"\n"+ s + s + s + s + "\n";
					// System.err.println("\n" + s + s + s + s);
					// System.err.println("The constraint " + constraint.getName() + " is not satisfied " + " \t==>  " + agId1 + "=" + val1 + " <> "+
					// agId2 + "=" + val2);
					// System.err.println(solution);
					// System.err.println(s + s + s + s + "\n");
					return problem;
				}
			}
		}
		return null;
	}

	/**
	 * Set the max number of non concurrent constraint checks (NCCCs) to the given value
	 * 
	 * @param cutOffNCCCs
	 *            the cut-off of NCCCs
	 */
	public void setCutOffNCCCs(long cutOffNCCCs) {
		NCCCObserver nccc = new NCCCObserver();
		nccc.setCutOffNCCCs(cutOffNCCCs);
		this.master.addMetric(nccc);
	}

	@Override
	public String getEndState() {
		return this.master.getEndState();
	}
	/**
	 * 
	 * @param agent
	 */
	public void setLocalAgents(AbstractAgent agent) {
		HashMap<Integer, AbstractAgent> agents = this.problem.getAgents();
		Iterator<Integer> it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = agents.get(it.next());
			if (!agent.description.equals("") && agent.description.equals(ag.description)) {
				agent.addLocalAgent(ag.getAgentId());
			}
		}
	}
	
	/**
	 * @param msg
	 */
	@Override
	public void updateAgentsLoad(Message msg) {
		int agId = msg.getSenderId();
		Long cc = (Long) msg.get(Commands.CCs);
		if(cc != null){
			String name = agId+"";//this.problem.getName(agId);
			this.CCs.put(name , cc);
			this.dataSetCCs.addValue(cc, name, name);
		}
	}
	/**
	 * @param msg
	 */
	public void setAgentsLoad(DefaultCategoryDataset dataset) {
		this.dataSetCCs = dataset;
	}
	
	/**
	 * @param msg
	 */
	public DefaultCategoryDataset getAgentsLoad() {
		return this.dataSetCCs;
	}

}
