package simulation.metric;

import java.util.ArrayList;
import java.util.Collection;

import kernel.AgentObserver;
import kernel.StopException;
import kernel.com.Message;
import kernel.events.AgentEvent;
import kernel.events.AgentFinishedEvent;
import util.Commands;
import util.Logger;

/**
 * A class for measuring the number of non concurrent constraint checks
 * 
 * @author Redouane EZZAHIR
 * 
 */
public class NCCCObserver extends AgentObserver implements MetricListener {

	/**
	 * The default cut-off value (1e7) of non-concurrent constraint checks
	 */

	protected long cutOffNCCCs = (long) 1e7;
	/**
	 * Current AgentC_Cost (NCCC) number of NCCCs of the observed agent
	 */
	protected long nbNCCCs = 0;
	/**
	 * The number of generic non-concurrent constraints checks
	 */
	public long nbGNCCCs = 0;
	public long nbCCs = 0;

	/**
	 * Constructs an observer for NCCCs
	 */
	public NCCCObserver() {
		super(Commands.NCCCs);
	}

	/**
	 * Updates the current Non-Concurrent constraint checks to be the maximum between its value and the value attached in the message <code>msg</code>
	 * 
	 * @throws StopException
	 */
	@Override
	public boolean handleReceivedMessage(Message msg) throws StopException {
		if (msg.contains(Commands.NCCCs)) {
			long senderCost = (Long) msg.get(Commands.NCCCs);
			this.nbNCCCs = Math.max(this.nbNCCCs, senderCost);
			//this.nbCCs += (Long) msg.get(Commands.CCs);
			if (this.nbNCCCs > this.cutOffNCCCs) {
				throw new StopException(Commands.ERROR);// + " NCCCs Cut-off");
			}
			return true;
		}
		return false;
	}

	/**
	 * Attach the Current number of NCCCs of the observed agent to the message <code>msg</code>
	 */
	@Override
	public void handleSendMessage(Message msg) {
		if (this.myAgent.isDisSearchStarted()) {
			msg.addContents(Commands.NCCCs, nbNCCCs);
			msg.addContents(Commands.GNCCCs, nbGNCCCs);
			msg.addContents(Commands.CCs, nbCCs);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void handleIncrementNbConstraintChecks(int n) {
		this.nbNCCCs += n;
		this.nbGNCCCs += n;
		this.nbCCs += n;
	}

	@Override
	public MetricListener clone() {
		MetricListener clone = null;
		try {
			clone = (NCCCObserver) super.clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return clone;

	}

	public void setCutOffNCCCs(long cutOffNCCCs) {
		this.cutOffNCCCs = cutOffNCCCs;
	}

	@Override
	public Collection<Metric> getEndStatistics() {
		ArrayList<Metric> m = new ArrayList<Metric>();
		m.add(new Metric(Commands.NCCCs, this.nbNCCCs, Metric.max));
		m.add(new Metric(Commands.GNCCCs, this.nbGNCCCs, Metric.max));
		m.add(new Metric(Commands.CCs, this.nbCCs, Metric.plus));
//		System.err.println(this.myAgent.agentId+" has performed "+this.nbCCs+" CCs");
		return m;
	}

	/**
	 * This function Handles the event <code> {@link kernel.events.AgentFinishedEvent} </code> It print in the logger the end value of
	 * AgentC_Cost (NCCC)
	 */
	@Override
	public void handleAgentEvent(AgentEvent event) {
		if (event instanceof AgentFinishedEvent) {
			this.myAgent.log("AgentC_Cost (#ncccs) : " + this.nbNCCCs, Logger.INFO_LEVEL);
			this.myAgent.log("AgentC_Cost (#gncccs) : " + this.nbGNCCCs, Logger.INFO_LEVEL);
			this.myAgent.log("AgentC_Cost (#CCs) : " + this.nbCCs, Logger.INFO_LEVEL);
		}
	}

	@Override
	public void handleIncrementCycle() {}

}
