package util.parser;

import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.TreeMap;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import gui.visualizing.Edge;
import gui.visualizing.Graph;
import gui.visualizing.Pair;
import gui.visualizing.Vertex;
import kernel.Constraint;
import kernel.order.AgentComplexity;
import util.Commands;
import util.parser.PredicateTokens.RelationalOperator;

/**
 * This class corresponds to a Java parser that uses DOM (Document Object Model) to parse DisCSP and WDisCSP instances in format "XDisCSP 1.0". <br>
 * Here, we assume that the instance is well-formed (valid). This class is given for illustration purpose. Feel free to adapt it ! Inspired from model
 * XCSP 2.1 proposed by christophe lecoutre
 * 
 * @version 1.0
 */
public class InstanceParser {

	public static final String VERSION = "version 1.0 (Janury 5, 2009)";
	private Document document = null;
	private String problemName = null;
	private String format = "XDisCSP 1.0";
	private String type = InstanceTokens.DisCSP;

	/**
	 * Simple or complex
	 */
	private String model = InstanceTokens.SIMPLE_MODEL;
	/**
	 * Constraint model : Totally (TKC) or Partial known (PKC) Constraints
	 */
	public static String constraintModel = InstanceTokens.TKC;

	public void setPKCModel() {
		constraintModel = InstanceTokens.PKC;
	}

	public void setComplexModel() {
		model = InstanceTokens.COMPLEX_MODEL;
	}

	public void setSimpleModel() {
		model = InstanceTokens.SIMPLE_MODEL;
	}

	private int maxConstraintArity = 0;
	private Map<String, PDomain> mapOfDomains = null;
	private Map<String, PAgent> mapOfAgents = null;
	private Map<String, PVariable> mapOfVariables = null;
	private Map<String, PRelation> mapOfRelations = null;
	private Map<String, PFunction> mapOfFunctions = null;
	private Map<String, PPredicate> mapOfPredicates = null;
	private Map<String, PConstraint> mapOfConstraints = null;
	private PVariable[] variables = null;
	private PAgent[] agents = null;
	private int nbExtensionConstraints = 0;
	private int nbSoftIntensionConstraints = 0;
	private int nbIntensionConstraints = 0;
	private int nbGlobalConstraints = 0;

	/**
	 * Used to determine if elements of the instance must be displayed when parsing.
	 */
	private boolean displayInstance = false;

	/**
	 * Used for WDisCSP
	 */
	private int maximalCost = 1000000;
	/**
	 * Used for WDisCSP
	 */
	private int initialCost = 1;

	private String satisfiable = "?";

	private String minViolatedConstraints = "?";

	public InstanceParser(String problemName) {
		this.mapOfAgents = new TreeMap<String, PAgent>();
		this.mapOfVariables = new TreeMap<String, PVariable>();
		this.mapOfDomains = new TreeMap<String, PDomain>();
		this.mapOfConstraints = new TreeMap<String, PConstraint>();
		this.mapOfPredicates = new TreeMap<String, PPredicate>();
		this.mapOfFunctions = new TreeMap<String, PFunction>();
		this.mapOfRelations = new TreeMap<String, PRelation>();
		this.problemName = problemName;
	}

	public String getType() {
		return this.type;
	}

	public String getModel() {
		return this.model;
	}

	public String getProblemName() {
		return this.problemName;
	}

	public void setProblemName(String name) {
		this.problemName = name;
	}

	public PVariable[] getVariables() {
		return this.variables;
	}

	public int getMaxConstraintArity() {
		return this.maxConstraintArity;
	}

	public Map<String, PConstraint> getMapOfConstraints() {
		return this.mapOfConstraints;
	}

	public Map<String, PRelation> getMapOfRelations() {
		return this.mapOfRelations;
	}

	public Map<String, PAgent> getMapOfAgents() {
		return this.mapOfAgents;
	}

	public Map<String, PVariable> getMapOfVariables() {
		return this.mapOfVariables;
	}

	public Map<String, PDomain> getMapOfDomains() {
		return this.mapOfDomains;
	}

	public Map<String, PPredicate> getMapOfPredicates() {
		return this.mapOfPredicates;
	}

	public Map<String, PFunction> getMapOfPFunctions() {
		return this.mapOfFunctions;
	}

	public int getNbExtensionConstraints() {
		return this.nbExtensionConstraints;
	}

	public int getNbIntensionConstraints() {
		return this.nbIntensionConstraints;
	}

	public int getNbGlobalConstraints() {
		return this.nbGlobalConstraints;
	}

	public String getConstraintsCategory() {
		return (nbExtensionConstraints > 0? "E" : "") + (nbIntensionConstraints > 0? "I" : "") + (nbGlobalConstraints > 0? "G" : "");
	}

	public String getSatisfiable() {
		return this.satisfiable;
	}

	public String getMinViolatedConstraints() {
		return this.minViolatedConstraints;
	}

	/**
	 * Build a DOM object that corresponds to the file whose name is given. <br>
	 * The file must represent a CSP instance according to format XCSP 2.1
	 * 
	 * @param fileName
	 *            the name of a file representing a CSP instance.
	 */
	public void loadInstance(String fileName) {
		this.document = XMLManager.load(fileName);
	}

	private void parsePresentation(Element presentationElement) {
		String s = presentationElement.getAttribute(InstanceTokens.MAX_CONSTRAINT_ARITY.trim());
		maxConstraintArity = s.length() == 0 || s.equals("?")? -1 : Integer.parseInt(s);
		this.type = presentationElement.getAttribute(InstanceTokens.TYPE.trim());
		this.type = this.type.length() == 0 || this.type.equals("?")? InstanceTokens.DisCSP : this.type;
		// DisCSP, WDisCSP or DCOP
		this.format = presentationElement.getAttribute(InstanceTokens.FORMAT.trim()); // XDisCSP 1.0
		this.problemName = presentationElement.getAttribute(InstanceTokens.NAME.trim());
		this.model = presentationElement.getAttribute(InstanceTokens.MODEL.trim());// simple or complex
		constraintModel = presentationElement.getAttribute(InstanceTokens.CONSTRAINT_MODEL.trim());// TKC or
		// PKC
		if (displayInstance)
			System.out.println(this.presentationXML());
		s = presentationElement.getAttribute(InstanceTokens.NB_SOLUTIONS).trim();
		satisfiable = s.length() == 0 || s.equals("?")? "unknown" : s.equals("0")? "false" : "true";
		s = presentationElement.getAttribute(InstanceTokens.MIN_VIOLATED_CONSTRAINTS).trim();
		minViolatedConstraints = satisfiable.equals("true")? "0" : s.length() == 0 || s.equals("?")? "unknown" : s;
	}

	public String presentationXML() {
		return "\t<presentation name=\"" + problemName + "\" " + InstanceTokens.MODEL + "=\"" + model + "\" " + InstanceTokens.CONSTRAINT_MODEL
				+ "=\"" + constraintModel + "\" format=\"" + format + "\" type=\"" + type + "\"/>  \n";
	}

	private int[] parseDomainValues(int nbValues, String stringOfValues) {
		int cnt = 0;
		int[] values = new int[nbValues];
		StringTokenizer st = new StringTokenizer(stringOfValues);
		while (st.hasMoreTokens()) {
			String token = st.nextToken();
			int position = token.indexOf(InstanceTokens.DISCRETE_INTERVAL_SEPARATOR);
			if (position == -1)
				values[cnt++] = Integer.parseInt(token);
			else {
				int min = Integer.parseInt(token.substring(0, position));
				int max = Integer.parseInt(token.substring(position + InstanceTokens.DISCRETE_INTERVAL_SEPARATOR.length()));
				for (int j = min ; j <= max ; j++)
					values[cnt++] = j;
			}
		}
		return values;
	}

	private PDomain parseDomain(Element domainElement) {
		String name = domainElement.getAttribute(InstanceTokens.NAME);
		int nbValues = Integer.parseInt(domainElement.getAttribute(InstanceTokens.NB_VALUES));
		int[] values = parseDomainValues(nbValues, domainElement.getTextContent());
		if (nbValues != values.length)
			throw new RuntimeException();
		return new PDomain(name, values);
	}

	private void parseDomains(Element domainsElement) {
		int nbDomains = Integer.parseInt(domainsElement.getAttribute(InstanceTokens.NB_DOMAINS));
		if (displayInstance)
			System.out.println("\t<domains nbDomains=\"" + nbDomains + "\">");
		NodeList nodeList = domainsElement.getElementsByTagName(InstanceTokens.DOMAIN);
		for (int i = 0 ; i < nodeList.getLength() ; i++) {
			PDomain domain = parseDomain((Element) nodeList.item(i));
			mapOfDomains.put(domain.getName(), domain);
			if (displayInstance)
				System.out.println("\t\t" + domain);
		}
		if (displayInstance)
			System.out.println("\t</domains>\n");
	}

	public String domainXML() {
		StringBuffer buf = new StringBuffer();
		buf.append("\t<domains nbDomains=\"" + this.mapOfDomains.size() + "\">\n");
		Iterator<PDomain> it = this.mapOfDomains.values().iterator();
		while (it.hasNext()) {
			PDomain domain = it.next();
			buf.append("\t\t" + domain.toString());
		}
		buf.append("\t</domains>\n");
		return buf.toString();
	}

	private PAgent parseAgent(Element agentElement) {
		String name = agentElement.getAttribute(InstanceTokens.NAME);
		int id = Integer.parseInt(agentElement.getAttribute(InstanceTokens.ID));
		String description = agentElement.getAttribute(InstanceTokens.DESCRIPTION);
		return new PAgent(name, id, description);
	}

	private void parseAgents(Element agentsElement) {
		int numAgents = Integer.parseInt(agentsElement.getAttribute(InstanceTokens.NB_AGENTS));
		if (displayInstance)
			System.out.println("\t<agents numAgents=\"" + numAgents + "\">\n");
		agents = new PAgent[numAgents];
		NodeList nodeList = agentsElement.getElementsByTagName(InstanceTokens.AGENT);
		for (int i = 0 ; i < nodeList.getLength() ; i++) {
			PAgent agent = parseAgent((Element) nodeList.item(i));
			mapOfAgents.put(agent.getName(), agent);
			agents[i] = agent;
			if (displayInstance)
				System.out.println("\t\t" + agent);
		}
		if (displayInstance)
			System.out.println("\t</agents>\n");
	}

	public String agentsXML() {
		StringBuffer buf = new StringBuffer();
		buf.append("\t<agents " + InstanceTokens.NB_AGENTS + "=\"" + this.mapOfAgents.size() + "\">\n");
		Iterator<PAgent> it = this.mapOfAgents.values().iterator();
		while (it.hasNext()) {
			PAgent agent = it.next();
			buf.append("\t\t" + agent.toString());
		}
		buf.append("\t</agents>\n");
		return buf.toString();
	}

	private PVariable parseVariable(Element variableElement) {
		String name = variableElement.getAttribute(InstanceTokens.NAME);
		int id = Integer.parseInt(variableElement.getAttribute(InstanceTokens.ID));
		String domainName = variableElement.getAttribute(InstanceTokens.DOMAIN);
		String description = variableElement.getAttribute(InstanceTokens.DESCRIPTION);
		String agentName = variableElement.getAttribute(InstanceTokens.AGENT);
		return new PVariable(mapOfAgents.get(agentName), name, id, description, mapOfDomains.get(domainName));
	}

	private void parseVariables(Element variablesElement) {
		int nbVariables = Integer.parseInt(variablesElement.getAttribute(InstanceTokens.NB_VARIABLES));
		if (displayInstance)
			System.out.println("\t<variables nbVariables=\"" + nbVariables + "\">\n");
		variables = new PVariable[nbVariables];
		NodeList nodeList = variablesElement.getElementsByTagName(InstanceTokens.VARIABLE);
		for (int i = 0 ; i < nodeList.getLength() ; i++) {
			PVariable variable = parseVariable((Element) nodeList.item(i));
			mapOfVariables.put(variable.getName(), variable);
			variables[i] = variable;
			if (displayInstance)
				System.out.println("\t\t" + variable.toString());
		}
		if (displayInstance)
			System.out.println("\t</variables>\n");
	}

	public String variablesXML() {
		StringBuffer buf = new StringBuffer();
		buf.append("\t<variables nbVariables=\"" + this.mapOfVariables.size() + "\">\n");
		Iterator<PVariable> it = this.mapOfVariables.values().iterator();
		while (it.hasNext()) {
			PVariable var = it.next();
			buf.append("\t\t" + var.toString());
		}
		buf.append("\t</variables>\n");
		return buf.toString();
	}

	private PRelation parseRelationTuples(String name, int arity, int nbTuples, String semantics, String textContent) {
		int[][] tuples = new int[nbTuples][arity];
		StringTokenizer st = new StringTokenizer(textContent, InstanceTokens.WHITE_SPACE + InstanceTokens.TUPLES_SEPARATOR);
		for (int i = 0 ; i < tuples.length ; i++)
			for (int j = 0 ; j < arity ; j++)
				tuples[i][j] = Integer.parseInt(st.nextToken());
		return new PRelation(name, arity, nbTuples, semantics, tuples);
	}

	private PRelation parseSoftRelationTuples(String name, int arity, int nbTuples, String semantics, String textContent, String textDefaultCost) {
		int[][] tuples = new int[nbTuples][arity];
		int[] weights = new int[nbTuples];
		StringTokenizer st = new StringTokenizer(textContent, InstanceTokens.WHITE_SPACE + InstanceTokens.TUPLES_SEPARATOR);
		int currentCost = -2;
		for (int i = 0 ; i < nbTuples ; i++) {
			String token = st.nextToken();
			int costFlagPosition = token.lastIndexOf(InstanceTokens.COST_SEPARATOR);
			if (costFlagPosition != -1) {
				currentCost = Integer.parseInt(token.substring(0, costFlagPosition));
				token = token.substring(costFlagPosition + 1);
			}
			weights[i] = currentCost;
			tuples[i][0] = Integer.parseInt(token);
			for (int j = 1 ; j < arity ; j++)
				tuples[i][j] = Integer.parseInt(st.nextToken());
		}
		int defaultCost = textDefaultCost.equals(InstanceTokens.INFINITY)? Constraint.MaxWeight : Integer.parseInt(textDefaultCost);
		return new PSoftRelation(name, arity, nbTuples, semantics, tuples, weights, defaultCost);
	}

	private PRelation parseRelation(Element relationElement) {
		String name = relationElement.getAttribute(InstanceTokens.NAME);
		int arity = Integer.parseInt(relationElement.getAttribute(InstanceTokens.ARITY));
		int nbTuples = Integer.parseInt(relationElement.getAttribute(InstanceTokens.NB_TUPLES));
		String semantics = relationElement.getAttribute(InstanceTokens.SEMANTICS);
		if (semantics.equals(InstanceTokens.SOFT))
			return parseSoftRelationTuples(name, arity, nbTuples, semantics, relationElement.getTextContent(), relationElement
					.getAttribute(InstanceTokens.DEFAULT_COST));
		else
			return parseRelationTuples(name, arity, nbTuples, semantics, relationElement.getTextContent());
	}

	private void parseRelations(Element relationsElement) {
		if (relationsElement == null)
			return;
		int nbRelations = Integer.parseInt(relationsElement.getAttribute(InstanceTokens.NB_RELATIONS));
		if (displayInstance)
			System.out.println("\t<relations nbRelations=\"" + nbRelations + "\">");
		NodeList nodeList = relationsElement.getElementsByTagName(InstanceTokens.RELATION);
		for (int i = 0 ; i < nodeList.getLength() ; i++) {
			PRelation relation = parseRelation((Element) nodeList.item(i));
			mapOfRelations.put(relation.getName(), relation);
			if (displayInstance)
				System.out.println("\t\t" + relation);
		}
		if (displayInstance)
			System.out.println("\t</relations>\n");
	}

	public String relationsXML() {
		StringBuffer buf = new StringBuffer();
		if (this.mapOfRelations.size() > 0) {
			buf.append("\t<relations nbRelations=\"" + this.mapOfRelations.size() + "\">\n");
			Iterator<PRelation> it = this.mapOfRelations.values().iterator();
			while (it.hasNext()) {
				PRelation r = it.next();
				buf.append("\t\t" + r.toString());
			}
			buf.append("\t</relations>\n");
		}
		return buf.toString();
	}

	private PPredicate parsePredicate(Element predicateElement) {
		String name = predicateElement.getAttribute(InstanceTokens.NAME);
		Element parameters = (Element) predicateElement.getElementsByTagName(InstanceTokens.PARAMETERS).item(0);
		Element expression = (Element) predicateElement.getElementsByTagName(InstanceTokens.EXPRESSION).item(0);
		Element functional = (Element) expression.getElementsByTagName(InstanceTokens.FUNCTIONAL).item(0);
		return new PPredicate(name, parameters.getTextContent().trim(), functional.getTextContent().trim());
	}

	private void parsePredicates(Element predicatesElement) {
		if (predicatesElement == null)
			return;
		int nbPredicates = Integer.parseInt(predicatesElement.getAttribute(InstanceTokens.NB_PREDICATES));
		if (displayInstance)
			System.out.println("\t<predicates nbPredicates=\"" + nbPredicates + "\">");
		NodeList nodeList = predicatesElement.getElementsByTagName(InstanceTokens.PREDICATE);
		for (int i = 0 ; i < nodeList.getLength() ; i++) {
			PPredicate predicate = parsePredicate((Element) nodeList.item(i));
			mapOfPredicates.put(predicate.getName(), predicate);
			if (displayInstance)
				System.out.println("\t\t" + predicate);
		}
		if (displayInstance)
			System.out.println("\t</predicates>\n");
	}

	public String predicatesXML() {
		StringBuffer buf = new StringBuffer();
		if (this.mapOfPredicates.size() > 0) {
			buf.append("\t<predicates nbPredicates=\"" + this.mapOfPredicates.size() + "\">\n");
			Iterator<PPredicate> it = this.mapOfPredicates.values().iterator();
			while (it.hasNext()) {
				PPredicate p = it.next();
				buf.append("\t\t" + p.toString());
			}
			buf.append("\t</predicates>\n");
		}
		return buf.toString();
	}

	private PFunction parseFunction(Element functionElement) {
		String name = functionElement.getAttribute(InstanceTokens.NAME);
		Element parameters = (Element) functionElement.getElementsByTagName(InstanceTokens.PARAMETERS).item(0);
		Element expression = (Element) functionElement.getElementsByTagName(InstanceTokens.EXPRESSION).item(0);
		Element functional = (Element) expression.getElementsByTagName(InstanceTokens.FUNCTIONAL).item(0);
		return new PFunction(name, parameters.getTextContent(), functional.getTextContent());
	}

	private void parseFunctions(Element functionsElement) {
		if (functionsElement == null)
			return;
		int nbFunctions = Integer.parseInt(functionsElement.getAttribute(InstanceTokens.NB_FUNCTIONS));
		if (displayInstance)
			System.out.println("\t<functions nbFunctions=\"" + nbFunctions + "\">");
		NodeList nodeList = functionsElement.getElementsByTagName(InstanceTokens.FUNCTION);
		for (int i = 0 ; i < nodeList.getLength() ; i++) {
			PFunction function = parseFunction((Element) nodeList.item(i));
			mapOfFunctions.put(function.getName(), function);
			if (displayInstance)
				System.out.println("\t\t" + function);
		}
		if (displayInstance)
			System.out.println("\t</functions>\n");
	}

	public String functionsXML() {
		StringBuffer buf = new StringBuffer();
		if (this.mapOfFunctions.size() > 0) {
			buf.append("\t<functions nbFunctions=\"" + this.mapOfFunctions.size() + "\">\n");
			Iterator<PFunction> it = this.mapOfFunctions.values().iterator();
			while (it.hasNext()) {
				PFunction r = it.next();
				buf.append("\t\t" + r.toString());
			}
			buf.append("\t</functions>\n");
		}
		return buf.toString();
	}

	private PVariable[] parseScope(String scope) {
		StringTokenizer st = new StringTokenizer(scope, " ");
		PVariable[] involvedVariables = new PVariable[st.countTokens()];
		for (int i = 0 ; i < involvedVariables.length ; i++)
			involvedVariables[i] = mapOfVariables.get(st.nextToken());
		return involvedVariables;
	}

	private int searchIn(String s, PVariable[] t) {
		for (int i = 0 ; i < t.length ; i++)
			if (t[i].getName().equals(s))
				return i;
		return -1;
	}

	private PConstraint parseElementConstraint(String name, PVariable[] scope, Element parameters) {
		StringTokenizer st =
				new StringTokenizer(Toolkit.insertWhitespaceAround(parameters.getTextContent().trim(), InstanceTokens.BRACKETS),
						InstanceTokens.WHITE_SPACE);
		PVariable index = mapOfVariables.get(st.nextToken()); // index is
		// necessarily a
		// variable
		st.nextToken(); // token [ skipped
		List<Object> table = new ArrayList<Object>();
		String token = st.nextToken();
		while (!token.equals("]")) {
			Object object = mapOfVariables.get(token);
			if (object == null)
				object = Integer.parseInt(token);
			table.add(object);
			token = st.nextToken();
		}
		token = st.nextToken();
		Object value = mapOfVariables.get(token);
		if (value == null)
			value = Integer.parseInt(token);
		return new PElement(name, scope, index, table.toArray(new Object[table.size()]), value);
	}

	private PConstraint parseWeightedSumConstraint(String name, PVariable[] scope, Element parameters) {
		NodeList nodeList = parameters.getChildNodes();
		StringTokenizer st = new StringTokenizer(nodeList.item(0).getTextContent(), InstanceTokens.WHITE_SPACE + "[{}]");
		int[] coeffs = new int[scope.length];
		while (st.hasMoreTokens()) {
			int coeff = Integer.parseInt(st.nextToken());
			int position = searchIn(st.nextToken(), scope);
			coeffs[position] += coeff;
		}
		RelationalOperator operator = RelationalOperator.getRelationalOperatorFor(nodeList.item(1).getNodeName());
		int limit = Integer.parseInt(nodeList.item(2).getTextContent().trim());
		return new PWeightedSum(name, scope, coeffs, operator, limit);
	}

	private String buildStringRepresentationOf(Element parameters) {
		NodeList nodeList = parameters.getChildNodes();
		StringBuffer sb = new StringBuffer();
		for (int i = 0 ; i < nodeList.getLength() ; i++) {
			Node node = nodeList.item(i);
			if (node.getNodeName().equals(InstanceTokens.NIL))
				sb.append(" ").append(InstanceTokens.NIL).append(" ");
			else
				sb.append(Toolkit.insertWhitespaceAround(node.getTextContent(), InstanceTokens.BRACKETS));
		}
		return sb.toString();
	}

	private PConstraint parseCumulativeConstraint(String name, PVariable[] scope, Element parameters) {
		StringTokenizer st = new StringTokenizer(buildStringRepresentationOf(parameters), InstanceTokens.WHITE_SPACE + "{}");
		st.nextToken(); // token '[' skipped
		List<Task> tasks = new ArrayList<Task>();
		String token = st.nextToken();
		while (!token.equals("]")) {
			Object origin = mapOfVariables.get(token);
			if (origin == null)
				origin = token.equals(InstanceTokens.NIL)? null : Integer.parseInt(token);
			token = st.nextToken();
			Object duration = mapOfVariables.get(token);
			if (duration == null)
				duration = token.equals(InstanceTokens.NIL)? null : Integer.parseInt(token);
			token = st.nextToken();
			Object end = mapOfVariables.get(token);
			if (end == null)
				end = token.equals(InstanceTokens.NIL)? null : Integer.parseInt(token);
			token = st.nextToken();
			Object height = mapOfVariables.get(token);
			if (height == null)
				height = Integer.parseInt(token);
			tasks.add(new Task(origin, duration, end, height));
			token = st.nextToken();
		}
		int limit = Integer.parseInt(st.nextToken());
		return new PCumulative(name, scope, tasks.toArray(new Task[tasks.size()]), limit);
	}

	private PConstraint parseConstraint(Element constraintElement) {
		PConstraint constr = null;
		String name = constraintElement.getAttribute(InstanceTokens.NAME);

		int arity = Integer.parseInt(constraintElement.getAttribute(InstanceTokens.ARITY));
		if (arity > maxConstraintArity)
			maxConstraintArity = arity;
		PVariable[] scope = parseScope(constraintElement.getAttribute(InstanceTokens.SCOPE));
		String reference = constraintElement.getAttribute(InstanceTokens.REFERENCE);
		if (mapOfRelations.containsKey(reference)) {
			nbExtensionConstraints++;
			constr = new PExtensionConstraint(name, scope, mapOfRelations.get(reference));
		} else if (mapOfFunctions.containsKey(reference)) {
			Element parameters = (Element) constraintElement.getElementsByTagName(InstanceTokens.PARAMETERS).item(0);
			nbSoftIntensionConstraints++;
			constr = new PIntensionConstraint(name, scope, mapOfFunctions.get(reference), parameters.getTextContent());
		} else if (mapOfPredicates.containsKey(reference)) {
			Element parameters = (Element) constraintElement.getElementsByTagName(InstanceTokens.PARAMETERS).item(0);
			nbIntensionConstraints++;
			constr = new PIntensionConstraint(name, scope, mapOfPredicates.get(reference), parameters.getTextContent());
		} else {
			nbGlobalConstraints++;
			String lreference = reference.toLowerCase();
			Element parameters = (Element) constraintElement.getElementsByTagName(InstanceTokens.PARAMETERS).item(0);

			if (lreference.equals(InstanceTokens.getLowerCaseGlobalNameOf(InstanceTokens.ALL_DIFFERENT)))
				constr = new PAllDifferent(name, scope);
			else if (lreference.equals(InstanceTokens.getLowerCaseGlobalNameOf(InstanceTokens.ELEMENT)))
				constr = parseElementConstraint(name, scope, parameters);
			else if (lreference.equals(InstanceTokens.getLowerCaseGlobalNameOf(InstanceTokens.WEIGHTED_SUM)))
				constr = parseWeightedSumConstraint(name, scope, parameters);
			else if (lreference.equals(InstanceTokens.getLowerCaseGlobalNameOf(InstanceTokens.CUMULATIVE)))
				constr = parseCumulativeConstraint(name, scope, parameters);
			else
				System.out.println("Problem with the reference " + reference);
		}
		if (constr != null && constraintModel.equals(Commands.PKC)) {
			int ownerId = Integer.parseInt(constraintElement.getAttribute(InstanceTokens.OWNER_ID));
			constr.setOwnerId(ownerId);
		}
		return constr;
	}

	private void parseConstraints(Element constraintsElement) {
		int nbConstraints = Integer.parseInt(constraintsElement.getAttribute(InstanceTokens.NB_CONSTRAINTS));
		if (displayInstance) {
			System.out.println("\t<constraints nbConstraints=\"" + nbConstraints + "\">");
			if (type.equals(InstanceTokens.WDisCSP)) {
				maximalCost = Integer.parseInt(constraintsElement.getAttribute(InstanceTokens.MAXIMAL_COST));
				String s = constraintsElement.getAttribute(InstanceTokens.INITIAL_COST);
				initialCost = s.equals("")? 0 : Integer.parseInt(s);
				System.out.print(" maximalCost=" + maximalCost + " initialCost=" + initialCost);
			}
		}

		NodeList nodeList = constraintsElement.getElementsByTagName(InstanceTokens.CONSTRAINT);
		for (int i = 0 ; i < nodeList.getLength() ; i++) {
			PConstraint constraint = parseConstraint((Element) nodeList.item(i));
			mapOfConstraints.put(constraint.getName(), constraint);
			if (displayInstance) {
				System.out.println("\t\t" + constraint);
			}
		}
		if (displayInstance)
			System.out.println("\t</constraints>\n");
	}

	public String constraintsXML() {
		StringBuffer buf = new StringBuffer();
		if (this.mapOfConstraints.size() > 0) {
			buf.append("\t<constraints nbConstraints=\"" + this.mapOfConstraints.size() + "\">\n");
			Iterator<PConstraint> it = this.mapOfConstraints.values().iterator();
			while (it.hasNext()) {
				PConstraint c = it.next();
				buf.append("\t\t" + c.toString());
			}
			buf.append("\t</constraints>\n");
		}
		return buf.toString();
	}

	/**
	 * Parse the DOM object that has been loaded.
	 * 
	 * @param displayInstance
	 *            if <code> true </code>, elements of the instance will be displayed.
	 */
	public void parse(boolean displayInstance) {
		this.displayInstance = displayInstance;
		parsePresentation((Element) document.getDocumentElement().getElementsByTagName(InstanceTokens.PRESENTATION).item(0));
		parseAgents((Element) document.getDocumentElement().getElementsByTagName(InstanceTokens.AGENTS).item(0));
		parseDomains((Element) document.getDocumentElement().getElementsByTagName(InstanceTokens.DOMAINS).item(0));
		parseVariables((Element) document.getDocumentElement().getElementsByTagName(InstanceTokens.VARIABLES).item(0));
		try {
			parseRelations((Element) document.getDocumentElement().getElementsByTagName(InstanceTokens.RELATIONS).item(0));
		} catch (NullPointerException e) {
			e.printStackTrace();
		}
		try {
			parseFunctions((Element) document.getDocumentElement().getElementsByTagName(InstanceTokens.FUNCTIONS).item(0));
		} catch (NullPointerException e) {
			e.printStackTrace();
		}
		try {
			parsePredicates((Element) document.getDocumentElement().getElementsByTagName(InstanceTokens.PREDICATES).item(0));
		} catch (NullPointerException e) {
			e.printStackTrace();
		}
		try {
			parseConstraints((Element) document.getDocumentElement().getElementsByTagName(InstanceTokens.CONSTRAINTS).item(0));
		} catch (NullPointerException e) {
			e.printStackTrace();
		}
	}

	public static void main(String[] args) {
		if (args.length != 1) {
			System.out.println("InstanceParser " + VERSION);
			System.out.println("Usage : java ... InstanceParser <instanceName>");
			System.exit(1);
		}

		InstanceParser parser = new InstanceParser(args[0]);
		parser.loadInstance(args[0]);
		parser.parse(true);
	}

	public void addPredicate(PPredicate predicate) {
		Collection<PPredicate> prd = this.mapOfPredicates.values();
		for (PPredicate pp : prd) {
			if (predicate.equals(pp)) {
				predicate.name = pp.name;
				return;
			}
		}
		this.mapOfPredicates.put(predicate.getName(), predicate);
	}

	public void addPredicate(PRelation relation) {
		Collection<PRelation> rls = this.mapOfRelations.values();
		for (PRelation r : rls) {
			if (relation.equals(r)) {
				relation.name = r.name;
				return;
			}
		}
		this.mapOfRelations.put(relation.getName(), relation);
	}

	public void addFunction(PFunction function) {
		Collection<PFunction> fct = this.mapOfFunctions.values();
		for (PFunction f : fct) {
			if (function.equals(f)) {
				function.name = f.name;
				return;
			}
		}
		this.mapOfFunctions.put(function.getName(), function);
	}

	public void addDomain(PDomain dom) {
		Iterator<PDomain> it = this.mapOfDomains.values().iterator();
		while (it.hasNext()) {
			PDomain d = it.next();
			if (d.equals(dom)) {
				dom.name = d.name;
				return;
			}
		}
		this.mapOfDomains.put(dom.getName(), dom);
	}

	public void addAgent(PAgent agent) {
		if (this.getMapOfAgents().containsKey(agent.getName())) {
			System.err.println("An agent with the same name already exist");
			System.exit(-1);
		}
		this.getMapOfAgents().put(agent.getName(), agent);
	}

	public void addVariable(PVariable var) {
		if (this.getMapOfVariables().containsKey(var.getName())) {
			System.err.println("An other variable  with the same name  already exist");
			System.exit(-1);
		}
		this.mapOfVariables.put(var.getName(), var);
	}

	public void addConstraint(PConstraint c) {
		if (!this.mapOfConstraints.containsKey(c.getName())) {
			this.mapOfConstraints.put(c.getName(), c);
		}
	}

	public String getFormat() {
		return this.format;
	}

	public void convertToSimpleModel() {
		if (this.model.equals(Commands.COMPLEX)) {
			this.model = Commands.SIMPLE;
			TreeMap<String, PAgent> newAgents = new TreeMap<String, PAgent>();
			Iterator<String> it = this.mapOfVariables.keySet().iterator();
			int i = 1;
			while (it.hasNext()) {
				PVariable var = this.mapOfVariables.get(it.next());
				PAgent ag = new PAgent(var.getName(), i, var.getAgent().getName());
				var.setAgent(ag);
				var.setId(0);
				newAgents.put(var.getName(), ag);
				i++;
			}
			this.mapOfAgents = null;
			this.mapOfAgents = newAgents;
		}
	}

	public void convertToCSPModel() {
		Iterator<String> it = this.mapOfVariables.keySet().iterator();
		HashMap<String, String> map = new HashMap<String, String>();
		int i = 1;
		while (it.hasNext()) {
			String key = it.next();
			PVariable var = this.mapOfVariables.get(key);
			String name = var.getName();
			String newName = name.substring(0, 1)+i;
			map.put(name, newName);
			var.setName(newName);
			var.setId(0);
			i++;
		}
		it = this.mapOfConstraints.keySet().iterator();
		while (it.hasNext()) {  
			String key = it.next();
			PConstraint c = this.mapOfConstraints.get(key);
			if(c instanceof PIntensionConstraint){
				String effective = ((PIntensionConstraint)c).effectiveParametersExpression;
				String[] names = effective.split(" ");
				((PIntensionConstraint)c).effectiveParametersExpression = "";
				for(String name : names){
					if(map.containsKey(name)){
						((PIntensionConstraint)c).effectiveParametersExpression += map.get(name)+" ";
					} else {
						((PIntensionConstraint)c).effectiveParametersExpression += name+" ";
					}
				}
			}
		}
	}

	public void initInstanceParser(String problemName) {
		this.mapOfAgents = null;
		this.mapOfVariables = null;
		this.mapOfDomains = null;
		this.mapOfConstraints = null;
		this.mapOfPredicates = null;
		this.mapOfFunctions = null;
		this.mapOfRelations = null;
		this.problemName = problemName;
		this.mapOfAgents = new TreeMap<String, PAgent>();
		this.mapOfVariables = new TreeMap<String, PVariable>();
		this.mapOfDomains = new TreeMap<String, PDomain>();
		this.mapOfConstraints = new TreeMap<String, PConstraint>();
		this.mapOfPredicates = new TreeMap<String, PPredicate>();
		this.mapOfFunctions = new TreeMap<String, PFunction>();
		this.mapOfRelations = new TreeMap<String, PRelation>();
		this.problemName = problemName;
	}

	public HashMap<Integer, AgentComplexity> getAgentsComplexity() {
		HashMap<Integer, AgentComplexity> complexities = new HashMap<Integer, AgentComplexity>();
		Iterator<String> it = this.mapOfVariables.keySet().iterator();
		while (it.hasNext()) {
			String key = it.next();
			PVariable variable = this.mapOfVariables.get(key);
			int agId = variable.getAgent().getId();
			if (!complexities.containsKey(agId)) {
				complexities.put(agId, new AgentComplexity());
			}
			complexities.get(agId).setMaxDomain(variable.getDomain().getNbValues());
			complexities.get(agId).IncrementNbLocalVars();
		}
		it = this.mapOfConstraints.keySet().iterator();
		while (it.hasNext()) {
			String key = it.next();
			PConstraint constraint = this.mapOfConstraints.get(key);
			PVariable[] scope = constraint.getScope();
			for (int v = 0 ; v < scope.length ; v++) {
				int agId = scope[v].getAgentId();
				complexities.get(agId).IncrementNbConstraints();
				for (int n = 0 ; n < scope.length ; n++) {
					if (v != n) {
						complexities.get(agId).addNeighbor(scope[n].getAgentId());
					}
				}
			}
		}
		return complexities;
	}

	public int getNbAgents() {
		return this.mapOfAgents.size();
	}

	public int getDomainSize() {
		Iterator<String> it = this.mapOfDomains.keySet().iterator();
		return this.mapOfDomains.get(it.next()).getNbValues();
	}

	public int getNbConstraints() {
		int nbConstraints = 0;
		nbConstraints += this.getNbExtensionConstraints();
		nbConstraints += this.getNbGlobalConstraints();
		nbConstraints += this.getNbIntensionConstraints();
		return nbConstraints;
	}

	public Document getDocument() {
		return this.document;
	}

	public int getNbVariables() {
		return this.mapOfVariables.size();
	}

	public void clear() {
		this.mapOfAgents.clear();
		this.mapOfConstraints.clear();
		this.mapOfDomains.clear();
		this.mapOfFunctions.clear();
		this.mapOfRelations.clear();
		this.mapOfRelations.clear();
		this.mapOfVariables.clear();
	}

	/**
	 * @return
	 */
	public Graph<Vertex, Edge> getGraph(java.awt.Dimension size) {
		Graph<Vertex, Edge> graph = new Graph<Vertex, Edge>();
		HashMap<Integer, Vertex> vertexes = new HashMap<Integer, Vertex>();
		for(String key : this.mapOfVariables.keySet()){
			int agId = this.mapOfVariables.get(key).getAgentId();
			double x = Math.random()*size.getWidth();
			double y = Math.random()*size.getHeight();
			vertexes.put(agId, new Vertex(agId, x, y));
		}
		for(String key : this.mapOfConstraints.keySet()){
			PConstraint constraint = this.mapOfConstraints.get(key);
			PVariable[] scope = constraint.getScope();
			Vertex v1 = vertexes.get(scope[0].getAgentId());
			Vertex v2 = vertexes.get(scope[1].getAgentId());
			graph.addEdge(new Edge(v1, v2), new Pair<Vertex>(v1, v2));
		}
		return graph;
	}

	public StringBuffer toStringBuffer(){
		StringBuffer buf = new StringBuffer();
		buf.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
		buf.append("<instance>\n");
		buf.append(this.presentationXML());
		buf.append(this.agentsXML());
		buf.append(this.domainXML());
		buf.append(this.variablesXML());
		buf.append(this.predicatesXML());
		buf.append(this.functionsXML());
		buf.append(this.relationsXML());
		buf.append(this.constraintsXML());
		buf.append("</instance>\n\n");
		return buf;
	}

	public void saveXML(String path, String name) {
		StringBuffer buf = this.toStringBuffer();
		try {
			File directory = new File(path + File.separator);
			if (!directory.exists()) { directory.mkdir(); }
			String file = directory.getPath() + File.separator + name;
			FileWriter writer = new FileWriter(file);
			writer.write(buf.toString());
			writer.close();
		} catch (java.io.IOException e) {
			e.printStackTrace();
		}

	}

	/**
	 * @param agId
	 * @return
	 */
	public String getAgentName(int agId) {
		for(String key : this.mapOfAgents.keySet()){
			PAgent agent = this.mapOfAgents.get(key);
			if(agent.getId() == agId){
				return agent.getName();
			}
		}
		return null;
	}
	

	/** 
	 * Returns the constraint graph in DOT format
	 * @return 	a String representation of the constraint graph in DOT format
	 */
	public String toDOT () {
		// Print the agents, with their respective variables
		StringBuilder out = new StringBuilder ("graph {\n\tnode [shape = \"circle\"];\n");
		for (PAgent agent : this.agents) {
			out.append("\tsubgraph cluster_" + agent.getName() + " {\n");
			out.append("\t\tlabel = " + agent.getName() + ";\n");
			for (PVariable var : this.variables) {
				if(var.getAgent().getId()==agent.getId()){
					String name = var.getName().replace(".", "_");
					out.append("\t\t" + name);
					out.append(" [style=\"filled\"]");
					// If var if a facade variable, fill it
//					for (PVariable neigh : this.getNeighborVars(var)) {
//						if (agent.getId() != neigh.getAgentId() ) {
//							out.append(" [style=\"filled\"]");
//							break;
//						}
//					}
					out.append(";\n");
				}
			}
			out.append("\t}\n");
		}
		out.append("\n");

		for(PConstraint constraint : this.mapOfConstraints.values()){
			PVariable[] scope = constraint.getScope();
			String var1 = scope[0].getName().replace(".", "_");
			String var2 = scope[1].getName().replace(".", "_");
			out.append("\t" + var1 + " -- " + var2 +" [label=\""+constraint.getName()+"\"] "+ ";\n");
		}
		out.append("}\n");
		return out.toString();
	}

	/**
	 * @return
	 */
	public boolean isSimpleModel() {
		return this.model.equals(InstanceTokens.SIMPLE_MODEL);
	}

	/**
	 * @return
	 */
	public boolean isComplexModel() {
		return this.model.equals(InstanceTokens.COMPLEX_MODEL);
	}

//	/**
//	 * @param var
//	 * @return
//	 */
//	private ArrayList<PVariable> getNeighborVars(PVariable var) {
//		ArrayList<PVariable> neighbors = new ArrayList<PVariable>();
//		for(String key : this.mapOfConstraints.keySet()){
//			PConstraint constraint = this.mapOfConstraints.get(key);
//			PVariable[] scope = constraint.getScope();
//			if(scope[0].getAgentId()==var.getAgentId() && scope[0].getId()==var.getId()){
//				neighbors.add(scope[1]);
//			} else if(scope[1].getAgentId()==var.getAgentId() && scope[1].getId()==var.getId()){
//				neighbors.add(scope[0]);
//			}
//		}
//		return neighbors;
//	}
}
